import platform
from typing import Dict

import pkg_resources
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

from cirro.api.auth.base import AuthInfo
from cirro.api.auth.iam import IAMAuth

try:
    sdk_version = pkg_resources.get_distribution("cirro").version
except (pkg_resources.RequirementParseError, TypeError):
    sdk_version = "Unknown"
python_version = platform.python_version()
headers = {
    'User-Agent': f'Cirro SDK {sdk_version} (Python {python_version})',
    'Accept': 'application/json',
    'Content-Type': 'application/json'
}


def _build_gql_client(auth_info: AuthInfo, endpoint: str):
    transport = RequestsHTTPTransport(url=endpoint, headers=headers, auth=auth_info.get_request_auth())
    return Client(transport=transport, fetch_schema_from_transport=False)


class ApiClient:
    def __init__(self, auth_info: AuthInfo, data_endpoint: str):
        self._auth_info = auth_info
        self._gql_client = _build_gql_client(auth_info, data_endpoint)

    def query(self, query: str, variables=None) -> Dict:
        return self._gql_client.execute(gql(query), variable_values=variables)

    @property
    def has_iam_creds(self) -> bool:
        return isinstance(self._auth_info, IAMAuth)

    @property
    def current_user(self) -> str:
        return self._auth_info.get_current_user()

    def get_iam_creds(self):
        if self.has_iam_creds:
            return self._auth_info.creds



from typing import Tuple, Dict

from cirro.api.auth import ClientAuth, UsernameAndPasswordAuth, IAMAuth
from cirro.cli.interactive.utils import ask, ask_yes_no


def gather_auth_config() -> Tuple[str, Dict]:
    auth_methods_map = {
        'Default authentication (recommended)': ClientAuth,
        'Username/password authentication (non-institutional)': UsernameAndPasswordAuth,
        'AWS IAM Credentials': IAMAuth
    }

    auth_method_answer = ask('select', 'Please select and authentication method',
                             choices=auth_methods_map.keys())

    auth_method = auth_methods_map[auth_method_answer]
    auth_method_config = {}

    if auth_method == UsernameAndPasswordAuth:
        auth_method_config['username'] = ask('text', 'Please enter your username', required=True)
        auth_method_config['password'] = ask('password', 'Please enter your password', required=True)

    if auth_method == ClientAuth:
        auth_method_config['enable_cache'] = ask_yes_no("Would you like to cache your login?")

    if auth_method == IAMAuth:
        if ask_yes_no("Would you like to manually specify AWS credentials?"):
            auth_method_config['access_key'] = ask('text', 'Please enter your access key ID', required=True)
            auth_method_config['secret_key'] = ask('text', 'Please enter your secret access key', required=True)
            auth_method_config['token'] = ask('text', 'Please enter your session token (optional)')
        else:
            auth_method_config['load_current'] = True

    return auth_method.__name__, auth_method_config



IGENOMES_REFERENCES = {
    "TAIR10": "Arabidopsis thaliana (TAIR10)",
    "EB2": "Bacillus subtilis 168 (EB2)",
    "UMD3.1": "Bos taurus (UMD3.1)",
    "bosTau8": "Bos taurus (bosTau8)",
    "WBcel235": "Caenorhabditis elegans (WBcel235)",
    "ce10": "Caenorhabditis elegans (ce10)",
    "CanFam3.1": "Canis familiaris (CanFam3.1)",
    "canFam3": "Canis familiaris (canFam3)",
    "GRCz10": "Danio rerio (GRCz10)",
    "danRer10": "Danio rerio (danRer10)",
    "BDGP6": "Drosophila melanogaster (BDGP6)",
    "dm6": "Drosophila melanogaster (dm6)",
    "EquCab2": "Equus caballus (EquCab2)",
    "equCab2": "Equus caballus (equCab2)",
    "EB1": "Escherichia coli K 12 DH10B (EB1)",
    "Galgal4": "Gallus gallus (Galgal4)",
    "Gm01": "Glycine max (Gm01)",
    "GRCh37": "Homo sapiens (GRCh37)",
    "GRCh38": "Homo sapiens (GRCh38)",
    "hg18": "Homo sapiens (hg18)",
    "hg19": "Homo sapiens (hg19)",
    "hg38": "Homo sapiens (hg38)",
    "Mmul 1": "Macaca mulatta (Mmul 1)",
    "GRCm38": "Mus musculus (GRCm38)",
    "mm10": "Mus musculus (mm10)",
    "IRGSP-1.0": "Oryza sativa japonica (IRGSP-1.0)",
    "CHIMP2.1.4": "Pan troglodytes (CHIMP2.1.4)",
    "panTro4": "Pan troglodytes (panTro4)",
    "Rnor 5.0": "Rattus norvegicus (Rnor 5.0)",
    "Rnor 6.0": "Rattus norvegicus (Rnor 6.0)",
    "rn6": "Rattus norvegicus (rn6)",
    "R64-1-1": "Saccharomyces cerevisiae (R64-1-1)",
    "sacCer3": "Saccharomyces cerevisiae (sacCer3)",
    "EF2": "Schizosaccharomyces pombe (EF2)",
    "Sbi1": "Sorghum bicolor (Sbi1)",
    "Sscrofa10.2": "Sus scrofa (Sscrofa10.2)",
    "susScr3": "Sus scrofa (susScr3)",
    "AGPv3": "Zea mays (AGPv3)"
}

PROCESSES_PATH_S3 = "s3://<RESOURCES_BUCKET>/process"



import json
import os
from collections import OrderedDict


class FormBuilder:
    """Build a form to drive inputs for running an analysis notebook non-interactively."""

    PARAM_TYPES = ['string', 'integer', 'boolean', 'number']

    def __init__(self):

        # Contents will be written out as the form
        self.form = dict(
            ui=OrderedDict(),
            form=OrderedDict(
                type="object",
                properties=OrderedDict()
            )
        )

        # Used to make sure that no keys are repeated
        self.used_keys = set()

        # Used to keep track when entering fields in a sub-section
        self.pointer = None

        # Store the params which will be populated either by:
        # a) While building the form, the optional `test_value` field will be used
        # b) While running non-interactively, it will use the values read from $PW_NOTEBOOK_DATA
        self.params = dict()

    def add_param(
        self,
        key: str = None,
        type: PARAM_TYPES = None,
        test_value=None,
        title: str = None,
        description: str = None,
        default=None,
        required: bool = None,
        multiple: bool = None
    ):
        """
        Add a parameter to the form.

        key:            (required) The unique identifier used to access the value provided by the user
        type:           (required) Variable type (options: string, integer, boolean, number)
        test_value:     (optional) The value which will be populated during interactive testing
        title:          (optional) Title describing the parameter
        description:    (optional) Longer description of the parameter
        default:        (optional) Default value shown to the user in the form
        required:       (optional) Whether or not the user must provide a value in the form
        multiple:       (optional) Whether or not the user may provide multiple values in the form
        """

        # Keys cannot be repeated
        assert key is not None, "Must specify key"
        assert key not in self.used_keys, f"Cannot repeat use of keys ('{key}')"
        self.used_keys.add(key)

        # Types must conform to the allowed values
        assert type is not None, "Must specify type"
        msg = f"Allowed values for type: '{', '.join(self.PARAM_TYPES)}', not '{type}'"
        assert type in self.PARAM_TYPES, msg

        # Start building the item in the form
        item = dict(type=type)

        # Populate the test value
        if test_value is not None:
            self.params[key] = test_value

        # Add the title
        if title is not None:
            item['title'] = title

        # Add the description
        if description is not None:
            item['description'] = description

        # Add the default
        if default is not None:
            item['default'] = default

        # Add the required
        if required is not None:
            item['required'] = required

        # Add the multiple
        if multiple is not None:
            item['multiple'] = multiple

        # Add the item to the form
        if self.pointer is None:
            self.form['form']['properties'][key] = item
        else:
            self.form['form']['properties'][self.pointer]['properties'][key] = item

    def add_section(self, title: str = None, description: str = None):
        """
        Add a section heading to the form.
        """

        # Name the section arbitrarily
        section_name = self._new_section_name()

        # Add new params to this section
        self.pointer = section_name

        # Add the section
        self.form['form']['properties'][section_name] = dict(
            title=title,
            description=description,
            properties=OrderedDict()
        )

    def _new_section_name(self):
        """Internal method to pick a new section name."""

        i = 1
        while f"section_{i}" in self.used_keys:
            i += 1
        self.used_keys.add(f"section_{i}")
        return f"section_{i}"

    def save(self):
        """
        When running interactively, save the form contents to notebook-form.json
        (or specify another path with $PW_NOTEBOOK_FORM).

        However, to support running non-interactively, if $PW_NOTEBOOK_DATA has been
        set to a local file, use those values to populate self.params instead.
        """

        # If running non-interactively, PW_NOTEBOOK_DATA will have been set
        if os.environ.get('PW_NOTEBOOK_DATA') is not None:

            notebook_data = os.environ.get('PW_NOTEBOOK_DATA')
            assert os.path.exists(notebook_data)

            self.params = {}

            # Read in the contents, and remove any section headings from the keys
            with open(notebook_data, 'r') as handle:
                for k, v in json.load(handle).items():
                    self.params[k.split(".")[-1]] = v

        # If running interactively, save out the form to notebook-form.json
        output_path = os.environ.get('PW_NOTEBOOK_FORM', 'notebook-form.json')
        with open(output_path, 'w') as handle:
            json.dump(self.form, handle, indent=4)



from abc import ABC
from typing import Dict

from cirro.api.clients import ApiClient
from cirro.api.config import AppConfig


def fetch_all_items(client: ApiClient, query: str, input_variables: Dict, batch_size=10000, max_items=None):
    """
    Fetches all items from a paginated graphql api
    """
    next_token = None
    items = []
    while True:
        variables = {
            'nextToken': next_token,
            'limit': batch_size,
            **input_variables
        }
        resp = client.query(query, variables)
        query_name = next(iter(resp.keys()))
        items.extend(resp[query_name]['items'])

        next_token = resp[query_name]['nextToken']
        if not next_token:
            return items

        if max_items and len(items) >= max_items:
            return items


class BaseService(ABC):
    _api_client: ApiClient
    _configuration: AppConfig

    def __init__(self, api_client: ApiClient, configuration: AppConfig):
        self._api_client = api_client
        self._configuration = configuration



from cirro.api.auth.base import AuthInfo
from cirro.api.auth.iam import IAMAuth
from cirro.api.auth.oauth_client import ClientAuth
from cirro.api.auth.username import UsernameAndPasswordAuth

__all__ = [
    'AuthInfo',
    'IAMAuth',
    'UsernameAndPasswordAuth',
    'ClientAuth',
    'get_auth_info_from_config'
]

from cirro.api.config import AppConfig


def get_auth_info_from_config(app_config: AppConfig):
    user_config = app_config.user_config
    if not user_config or not user_config.auth_method:
        return ClientAuth(region=app_config.region,
                          client_id=app_config.client_id,
                          auth_endpoint=app_config.auth_endpoint)

    auth_methods = [
        ClientAuth,
        UsernameAndPasswordAuth,
        IAMAuth
    ]
    matched_auth_method = next((m for m in auth_methods if m.__name__ == user_config.auth_method), None)
    if not matched_auth_method:
        raise RuntimeError(f'{user_config.auth_method} not found, please re-run configuration')

    auth_config = user_config.auth_method_config

    if matched_auth_method == ClientAuth:
        return ClientAuth(region=app_config.region,
                          client_id=app_config.client_id,
                          auth_endpoint=app_config.auth_endpoint,
                          enable_cache=auth_config.get('enable_cache') == 'True')

    if matched_auth_method == IAMAuth and auth_config.get('load_current'):
        return IAMAuth.load_current()

    if matched_auth_method == IAMAuth:
        return IAMAuth(region=app_config.region, **auth_config)

    if matched_auth_method == UsernameAndPasswordAuth:
        return UsernameAndPasswordAuth(region=app_config.region,
                                       client_id=app_config.client_id,
                                       user_pool_id=app_config.user_pool_id,
                                       **auth_config)



from cirro.cli.controller import run_ingest, run_download, run_configure, run_list_datasets, run_configure_workflow

__all__ = [
    'run_ingest',
    'run_download',
    'run_configure',
    'run_list_datasets',
    'run_configure_workflow'
]



from pathlib import Path
from typing import List, Optional

from cirro.api.models.file import FileAccessContext, CheckDataTypesInput
from cirro.api.models.form_specification import ParameterSpecification
from cirro.api.models.process import Executor, RunAnalysisCommand, Process
from cirro.api.models.s3_path import S3Path
from cirro.api.services.file import FileEnabledService


class ProcessService(FileEnabledService):
    def list(self, process_type: Executor = None) -> List[Process]:
        """
        Gets a list of processes filtered by an optional process type
        :param process_type:
        :return:
        """
        query = '''
          query ListProcesses(
            $filter: ModelProcessFilterInput
            $limit: Int
            $nextToken: String
          ) {
            listProcesses(filter: $filter, limit: $limit, nextToken: $nextToken) {
              items {
                id
                childProcessIds
                name
                desc
                executor
                documentationUrl
                code {
                  repository
                  uri
                  script
                  version
                }
                paramMapJson
                formJson
                fileRequirementsMessage
                fileMappingRules {
                  min
                  max
                  description
                  isSample
                  fileNamePatterns {
                    exampleName
                    description
                    sampleMatchingPattern
                  }
                }
              }
            }
          }
        '''
        item_filter = {}
        if process_type:
            item_filter['executor'] = {'eq': process_type.value}
        resp = self._api_client.query(query, variables={'filter': item_filter})['listProcesses']
        return [Process.from_record(p) for p in resp['items']]

    def get_process(self, process_id: str) -> Process:
        """
        Gets detail of the specified process
        """
        query = '''
          query GetProcess($id: ID!) {
            getProcess(id: $id) {
              id
              childProcessIds
              name
              desc
              executor
              documentationUrl
              code {
                repository
                uri
                script
                version
              }
              paramMapJson
              formJson
              fileRequirementsMessage
              fileMappingRules {
                min
                max
                description
                isSample
                fileNamePatterns {
                  exampleName
                  description
                  sampleMatchingPattern
                }
              }
            }
          }
        '''
        return Process.from_record(self._api_client.query(query, variables={'id': process_id})['getProcess'])

    def find_by_name(self, name: str) -> Optional[Process]:
        """
        Get a process by its display name
        """
        matched_process = next((p for p in self.list() if p.name == name), None)
        if not matched_process:
            return None

        return self.get_process(matched_process.id)

    def run_analysis(self, run_analysis_command: RunAnalysisCommand) -> str:
        """
        Runs an analysis on a dataset, returns the ID of the new dataset
        """
        spec = self.get_parameter_spec(run_analysis_command.process_id)
        spec.validate_params(run_analysis_command.params)
        query = '''
          mutation RunAnalysis($input: RunAnalysisInput!) {
            runAnalysis(input: $input)
          }
        '''
        resp = self._api_client.query(query, variables={'input': run_analysis_command.to_json()})
        return resp['runAnalysis']

    def get_parameter_spec(self, process_id: str) -> ParameterSpecification:
        """
        Gets a specification used to describe the parameters used in the process
        """
        access_context = FileAccessContext.resources(self._configuration.resources_bucket)
        process = self.get_process(process_id)
        path = S3Path(process.form_spec_json)
        if path.valid:
            form_spec_json = self._file_service.get_file_from_path(access_context, path.key)
        else:
            form_spec_json = process.form_spec_json
        return ParameterSpecification.from_json(form_spec_json)

    def check_dataset_files(self, files: List[str], process_id: str, directory: str):
        """
        Checks if the file mapping rules for a process are met by the list of files
        :param files: files to check
        :param process_id: ID for the process containing the file mapping rules
        :param directory: path to directory containing files
        """
        # Parse samplesheet file if present
        samplesheet = None
        samplesheet_file = Path(directory, 'samplesheet.csv')
        if samplesheet_file.exists():
            samplesheet = samplesheet_file.read_text()

        # Call cirro function
        data_types_input = CheckDataTypesInput(fileNames=files, processId=process_id, sampleSheet=samplesheet)
        query = '''
            query checkDataTypes($input: CheckDataTypesInput!) {
              checkDataTypes(input: $input) {
                files
                errorMsg
                allowedDataTypes {
                  description
                  errorMsg
                  allowedPatterns {
                    exampleName
                    description
                    sampleMatchingPattern
                  }
                }
              }
            }
        '''
        resp = self._api_client.query(query, variables={'input': data_types_input})
        reqs = resp['checkDataTypes']

        # These will be samplesheet errors or no files errors
        if reqs['errorMsg']:
            raise ValueError(reqs['errorMsg'])

        # These will be errors for missing files
        all_errors = [entry['errorMsg'] for entry in reqs['allowedDataTypes'] if entry['errorMsg'] is not None]
        patterns = [' or '.join([e['exampleName'] for e in entry['allowedPatterns']])
                    for entry in reqs['allowedDataTypes']]

        if len(all_errors) != 0:
            raise ValueError("Files do not meet dataset type requirements. The expected files are: \n" +
                             "\n".join(patterns))



from time import sleep
from typing import Union

from cirro.api.clients.portal import DataPortalClient
from cirro.api.models.dataset import CreateIngestDatasetInput
from cirro.api.models.project import Project
from cirro.file_utils import get_files_in_directory
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.dataset import DataPortalDataset, DataPortalDatasets
from cirro.sdk.exceptions import DataPortalAssetNotFound, DataPortalInputError
from cirro.sdk.helpers import parse_process_name_or_id
from cirro.sdk.process import DataPortalProcess
from cirro.sdk.reference import DataPortalReference, DataPortalReferences
from cirro.sdk.reference_type import DataPortalReferenceType, DataPortalReferenceTypes


class DataPortalProject(DataPortalAsset):
    """
    Projects in the Data Portal contain collections of Datasets.
    Users are granted permissions at the project-level, allowing them
    to view and/or modify all of the datasets in that collection.
    """
    name = None

    def __init__(self, proj: Project, client: DataPortalClient):
        """Initialize the Project from the base Cirro model."""

        self.id = proj.id
        self.name = proj.name
        self.description = proj.description
        self._client = client

    def __str__(self):
        """Control how the Project is rendered as a string."""

        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description']
        ])

    def list_datasets(self) -> DataPortalDatasets:
        """List all of the datasets available in the project."""

        return DataPortalDatasets(
            [
                DataPortalDataset(d, self._client)
                for d in self._client.dataset.find_by_project(self.id)
            ]
        )

    def get_dataset_by_name(self, name: str = None) -> DataPortalDataset:
        """Return the dataset with the specified name."""

        dataset = next(iter(self._client.dataset.find_by_project(self.id, name=name)), None)
        if dataset is None:
            raise DataPortalAssetNotFound(f'Dataset with name {name} not found')
        return DataPortalDataset(dataset, self._client)

    def get_dataset_by_id(self, _id: str = None) -> DataPortalDataset:
        """Return the dataset with the specified id."""

        dataset = self._client.dataset.get_from_id(_id=_id)
        if dataset is None:
            raise DataPortalAssetNotFound(f'Dataset with ID {_id} not found')
        return DataPortalDataset(dataset, self._client)

    def list_references(self, reference_type: str = None) -> DataPortalReferences:
        """
        List the references available in a project.
        Optionally filter to references of a particular type (identified by name)
        """

        # Get the complete list of references which are available
        references = DataPortalReferenceTypes(
            [
                DataPortalReferenceType(ref)
                for ref in self._client.common.get_references_types()
            ]
        )

        # If a particular name was specified
        if reference_type is not None:
            references = references.filter_by_pattern(reference_type)
            if len(references) == 0:
                msg = f"Could not find any reference types with the name {reference_type}"
                raise DataPortalAssetNotFound(msg)

        return DataPortalReferences(
            [
                DataPortalReference(ref)
                for reference_type in references
                for ref in self._client.project.get_references(
                    self.id,
                    reference_type.directory
                )
            ]
        )

    def get_reference_by_name(self, name: str = None, ref_type: str = None) -> DataPortalReference:
        """Return the reference of a particular type with the specified name."""

        if name is None:
            raise DataPortalInputError("Must specify the reference name")

        return self.list_references(ref_type).get_by_name(name)

    def upload_dataset(
        self,
        name: str = None,
        description='',
        process: Union[DataPortalProcess, str] = None,
        upload_folder: str = None,
        files: list = None
    ):
        """
        Upload a set of files to the Data Portal, creating a new dataset.

        If the files parameter is not provided, it will upload all files in the upload folder
        """

        if name is None:
            raise DataPortalInputError("Must provide name for new dataset")
        if process is None:
            raise DataPortalInputError("Must provide the process which is used for ingest")
        if upload_folder is None:
            raise DataPortalInputError("Must provide upload_folder -- folder containing files to upload")

        # Parse the process provided by the user
        process = parse_process_name_or_id(process, self._client)

        # If no files were provided
        if files is None:

            # Get the list of files in the upload folder
            files = get_files_in_directory(upload_folder)

        # Make sure that the files match the expected pattern
        self._client.process.check_dataset_files(files, process.id, upload_folder)

        # Create the ingest process request
        dataset_create_request = CreateIngestDatasetInput(
            project_id=self.id,
            process_id=process.id,
            name=name,
            description=description,
            files=files
        )

        # Get the response
        create_response = self._client.dataset.create(dataset_create_request)

        # Upload the files
        self._client.dataset.upload_files(
            project_id=self.id,
            dataset_id=create_response['datasetId'],
            directory=upload_folder,
            files=dataset_create_request.files
        )

        # Return the dataset which was created, which might take a second to update
        max_attempts = 5
        for attempt in range(max_attempts):
            try:
                return self.get_dataset_by_id(create_response['datasetId'])
            except DataPortalAssetNotFound as e:
                if attempt == max_attempts - 1:
                    raise e
                else:
                    sleep(2)


class DataPortalProjects(DataPortalAssets[DataPortalProject]):
    asset_name = "project"



import json
from dataclasses import dataclass
from enum import Enum
from typing import NamedTuple, Dict, Any, TypedDict, List, Optional

from cirro.api.models.exceptions import DataPortalModelException


class Executor(Enum):
    UNKNOWN = 'UNKNOWN'
    INGEST = 'INGEST'
    "Process type used when manually uploading files"
    NEXTFLOW = 'NEXTFLOW'
    "Processes that are ran using Nextflow"


@dataclass(frozen=True)
class ProcessCode:
    repository: str
    uri: str
    script: str
    version: str

    @classmethod
    def from_record(cls, record: Dict):
        if record is None:
            raise DataPortalModelException("Cannot construct Dataset from null object.")
        if not record:
            return None
        return cls(**record)


@dataclass(frozen=True)
class ProcessCompute:
    executor: str
    json: str
    name: str


@dataclass(frozen=True)
class Process:
    id: str
    name: str
    description: str
    child_process_ids: List[str]
    executor: Executor
    documentation_url: str
    code: ProcessCode
    form_spec_json: str
    sample_sheet_path: str
    file_requirements_message: str
    file_mapping_rules: List

    @classmethod
    def from_record(cls, record: 'ProcessRecord'):
        if record is None:
            raise DataPortalModelException("Cannot construct Dataset from null object.")
        return cls(
            record.get('id'),
            record.get('name'),
            record.get('desc'),
            record.get('childProcessIds'),
            Executor[record.get('executor')],
            record.get('documentationUrl'),
            ProcessCode.from_record(record.get('code', {})),
            record.get('formJson'),
            record.get('sampleSheetPath'),
            record.get('fileRequirementsMessage'),
            record.get('fileMappingRules'))


class ProcessRecord(TypedDict):
    id: str
    childProcessIds: List[str]
    name: str
    desc: str
    executor: str
    documentationUrl: str
    code: Dict
    paramDefaults: List
    computeDefaults: List[Dict]
    paramMapJson: str
    formJson: str
    fileJson: str
    componentJson: str
    infoJson: str
    webOptimizationJson: str
    preProcessScript: str
    sampleSheetPath: Optional[str]
    fileRequirementsMessage: str
    fileMappingRules: List


class RunAnalysisCommand(NamedTuple):
    name: str
    description: str
    process_id: str
    parent_dataset_id: str
    project_id: str
    params: Dict[str, Any]
    notifications_emails: List[str]

    def to_json(self):
        try:
            param_json = json.dumps(self.params)
        except TypeError as e:
            raise RuntimeError('Params invalid') from e

        return {
            'name': self.name,
            'description': self.description,
            'processId': self.process_id,
            'parentDatasetId': self.parent_dataset_id,
            'projectId': self.project_id,
            'paramJson': param_json,
            'notificationEmails': self.notifications_emails
        }



import json
from functools import partial
from typing import List

from cirro.api.clients import ApiClient, S3Client
from cirro.api.config import AppConfig
from cirro.api.models.auth import Creds
from cirro.api.models.file import FileAccessContext, File
from cirro.api.services.base import BaseService
from cirro.file_utils import upload_directory, download_directory


class FileService(BaseService):
    def get_access_credentials(self, access_context: FileAccessContext) -> Creds:
        # Special case:
        # we do not need to call the API to get IAM creds if we are using IAM creds
        if self._api_client.has_iam_creds:
            return self._api_client.get_iam_creds()
        # Call API to get temporary credentials
        credentials_response = self._api_client.query(*access_context.get_token_query)
        return credentials_response['getFileAccessToken']

    def get_file(self, file: File) -> bytes:
        """
        Gets the contents of a file
        """
        return self.get_file_from_path(file.access_context, file.relative_path)

    def get_file_from_path(self, access_context: FileAccessContext, file_path: str) -> bytes:
        """
        Gets the contents of a file by providing the path, used internally
        """
        s3_client = S3Client(partial(self.get_access_credentials, access_context))
        full_path = f'{access_context.path_prefix}/{file_path}'.lstrip('/')
        return s3_client.get_file(access_context.bucket, full_path)

    def create_file(self, access_context: FileAccessContext, key: str,
                    contents: str, content_type: str):
        """
        Creates a file at the specified path
        """
        s3_client = S3Client(partial(self.get_access_credentials, access_context))
        s3_client.create_object(key=key,
                                contents=contents,
                                content_type=content_type,
                                bucket=access_context.bucket)

    def upload_files(self, access_context: FileAccessContext, directory: str, files: List[str]):
        """
        Uploads a list of files from the specified directory
        :param access_context: File access context, use class methods to generate
        :param directory: base path to upload from
        :param files: relative path of files to upload
        :return:
        """
        s3_client = S3Client(partial(self.get_access_credentials, access_context))
        upload_directory(directory, files, s3_client, access_context.bucket, access_context.path_prefix)

    def download_files(self, access_context: FileAccessContext, directory: str, files: List[str]):
        """
        Download a list of files to the specified directory
        :param access_context: File access context, use class methods to generate
        :param directory: download location
        :param files: relative path of files to download
        """
        s3_client = S3Client(partial(self.get_access_credentials, access_context))
        download_directory(directory, files, s3_client, access_context.bucket, access_context.path_prefix)

    def get_file_listing(self, access_context: FileAccessContext) -> List[File]:
        """
        Gets a listing of files of the current access context,
        note that this expects a manifest.json file
        :param access_context: File access context, use class methods to generate
        :return: relative path of files
        """
        file = self.get_file_from_path(access_context, 'web/manifest.json')
        manifest = json.loads(file)
        return [File(file['file'], file['size'], access_context)
                for file in manifest['files']]


class FileEnabledService(BaseService):
    """
    Not to be instantiated directly
    """
    _file_service: FileService

    def __init__(self, api_client: ApiClient, file_service: FileService, configuration: AppConfig):
        super().__init__(api_client, configuration)
        self._file_service = file_service



from pathlib import Path, PurePath
from typing import List, Union

from boto3.exceptions import S3UploadFailedError

from cirro.api.clients import S3Client
from cirro.api.models.file import DirectoryStatistics, File

DEFAULT_TRANSFER_SPEED = 160


def filter_files_by_pattern(files: Union[List[File], List[str]], pattern: str) -> List[File]:
    """
    Filters a list of files by a glob pattern
    """
    def matches_glob(file: Union[File, str]):
        return PurePath(file if isinstance(file, str) else file.relative_path).match(pattern)

    return [
        file for file in files
        if matches_glob(file)
    ]


def get_files_in_directory(directory) -> List[str]:
    path = Path(directory)
    path_posix = str(path.as_posix())

    paths = []

    for file_path in path.rglob("*"):
        if file_path.is_dir():
            continue
        if file_path.name.startswith('.'):
            continue

        str_file_path = str(file_path.as_posix())
        str_file_path = str_file_path.replace(f'{path_posix}/', "")
        paths.append(str_file_path)

    return paths


def get_directory_stats(directory) -> DirectoryStatistics:
    sizes = [f.stat().st_size for f in Path(directory).glob('**/*') if f.is_file()]
    total_size = sum(sizes) / float(1 << 30)
    return {
        'sizeFriendly': f'{total_size:,.3f} GB',
        'size': total_size,
        'numberOfFiles': len(sizes)
    }


def upload_directory(directory: str, files: List[str], s3_client: S3Client, bucket: str, prefix: str, max_retries=10):
    for file in files:
        key = f'{prefix}/{file}'
        local_path = Path(directory, file)
        success = False

        # Retry up to max_retries times
        for retry in range(max_retries):

            # Try the upload
            try:
                s3_client.upload_file(
                    local_path=local_path,
                    bucket=bucket,
                    key=key
                )

                success = True

            # Catch the upload error
            except S3UploadFailedError as e:

                # Report the error
                print(f"Encountered error:\n{str(e)}\n"
                      f"Retrying ({max_retries - (retry + 1)} attempts remaining)")

            if success:
                break


def download_directory(directory: str, files: List[str], s3_client: S3Client, bucket: str, prefix: str):
    for file in files:
        key = f'{prefix}/{file}'
        local_path = Path(directory, file)
        local_path.parent.mkdir(parents=True, exist_ok=True)

        s3_client.download_file(local_path=local_path,
                                bucket=bucket,
                                key=key)


def estimate_token_lifetime(data_size_gb: float, speed_mbps: float = DEFAULT_TRANSFER_SPEED) -> int:
    """
    :param data_size_gb: Gigabytes
    :param speed_mbps: Megabits per second
    """
    transfer_time_seconds = (data_size_gb * 8 * 1000) / speed_mbps
    transfer_time_hours = transfer_time_seconds / 60 / 60
    return max(round(transfer_time_hours), 1)



import math
import threading
from pathlib import Path
from typing import Callable

from boto3 import Session
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from tqdm import tqdm

from cirro.api.models.auth import Creds
from cirro.utils import parse_json_date


def convert_size(size):
    if size == 0:
        return '0B'
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size, 1024)))
    p = math.pow(1024, i)
    s = round(size/p, 2)
    return '%.2f %s' % (s, size_name[i])


def format_creds_for_session(creds: Creds):
    expiration = parse_json_date(creds['Expiration'])
    return {
        'access_key': creds['AccessKeyId'],
        'secret_key': creds['SecretAccessKey'],
        'token': creds['SessionToken'],
        'expiry_time': expiration.isoformat()
    }


class ProgressPercentage:
    def __init__(self, progress: tqdm):
        self._lock = threading.Lock()
        self.progress = progress

    def __call__(self, bytes_amount):
        with self._lock:
            self.progress.update(bytes_amount)


class S3Client:
    def __init__(self, creds_getter: Callable[[], Creds]):
        self._creds_getter = creds_getter
        self._client = self._build_session_client()

    def upload_file(self, local_path: Path, bucket: str, key: str):
        file_size = local_path.stat().st_size
        file_name = local_path.name

        with tqdm(total=file_size,
                  desc=f'Uploading file {file_name} ({convert_size(file_size)})',
                  bar_format="{desc} | {percentage:.1f}%|{bar:25} | {rate_fmt}",
                  unit='B', unit_scale=True,
                  unit_divisor=1024) as progress:
            absolute_path = str(local_path.absolute())
            self._client.upload_file(absolute_path, bucket, key, Callback=ProgressPercentage(progress))

    def download_file(self, local_path: Path, bucket: str, key: str):
        file_size = self.get_file_stats(bucket, key)['ContentLength']
        file_name = local_path.name

        with tqdm(total=file_size,
                  desc=f'Downloading file {file_name} ({convert_size(file_size)})',
                  bar_format="{desc} | {percentage:.1f}%|{bar:25} | {rate_fmt}",
                  unit='B', unit_scale=True,
                  unit_divisor=1024) as progress:
            absolute_path = str(local_path.absolute())
            self._client.download_file(bucket, key, absolute_path, Callback=ProgressPercentage(progress))

    def create_object(self, bucket: str, key: str, contents: str, content_type: str):
        self._client.put_object(
            Bucket=bucket,
            Key=key,
            ContentType=content_type,
            ContentEncoding='utf-8',
            Body=bytes(contents, "UTF-8")
        )

    def get_file(self, bucket: str, key: str) -> bytes:
        resp = self._client.get_object(Bucket=bucket, Key=key)
        file_body = resp['Body']
        return file_body.read()

    def get_file_stats(self, bucket: str, key: str):
        return self._client.head_object(Bucket=bucket, Key=key)

    def _build_session_client(self):
        creds = self._creds_getter()

        if creds['Expiration']:
            session = get_session()
            session._credentials = RefreshableCredentials.create_from_metadata(
                metadata=format_creds_for_session(creds),
                refresh_using=self._refresh_credentials,
                method='sts'
            )
            session = Session(botocore_session=session)
        else:
            session = Session(
                aws_access_key_id=creds['AccessKeyId'],
                aws_secret_access_key=creds['SecretAccessKey'],
                aws_session_token=creds['SessionToken']
            )
        return session.client('s3')

    def _refresh_credentials(self):
        new_creds = self._creds_getter()
        return format_creds_for_session(new_creds)



import json
import logging
import uuid
from typing import List, Union, Optional

from cirro.api.clients.utils import filter_deleted
from cirro.api.models.dataset import CreateIngestDatasetInput, DatasetCreateResponse, Dataset
from cirro.api.models.file import FileAccessContext, File
from cirro.api.models.status import Status
from cirro.api.services.base import fetch_all_items
from cirro.api.services.file import FileEnabledService

logger = logging.getLogger()


class DatasetService(FileEnabledService):
    def find_by_project(self, project_id: str, name: str = None) -> List[Dataset]:
        """
         Lists datasets by project with an optional name provided
         Dataset names are not unique so providing a name doesn't guarantee one result
        """
        query = '''
          query DatasetsByProject(
            $project: ID!
            $sortDirection: ModelSortDirection
            $filter: ModelDatasetFilterInput
            $limit: Int
            $nextToken: String
          ) {
            datasetsByProject(
              project: $project
              sortDirection: $sortDirection
              filter: $filter
              limit: $limit
              nextToken: $nextToken
            ) {
              items {
                id
                project
                status
                name
                desc
                sourceDatasets
                infoJson
                paramJson
                process
                createdAt
                updatedAt
              }
              nextToken
            }
          }
        '''
        variables = {
            'project': project_id,
            'filter': {
                'status': {
                    'eq': Status.COMPLETED.value
                }
            }
        }
        if name:
            variables['filter']['name'] = {'eq': name}

        items = fetch_all_items(self._api_client, query, variables)
        not_deleted = filter_deleted(items)
        return [Dataset.from_record(item) for item in not_deleted]

    def get_from_id(self, _id: str) -> Optional[Dataset]:
        query = '''
             query GetDataset($id: ID!) {
                getDataset(id: $id) {
                  id
                  status
                  name
                  desc
                  sourceDatasets
                  paramJson
                  infoJson
                  process
                  project
                  createdAt
                  updatedAt
                }
              }
              '''

        item = self._api_client.query(query, variables={'id': _id})['getDataset']
        if not item:
            return

        return Dataset.from_record(item)

    def create(self, create_request: CreateIngestDatasetInput) -> DatasetCreateResponse:
        """
        Creates an ingested dataset.
        This only registers into the system, does not upload any files
        """
        logger.info(f"Creating dataset {create_request.name}")

        if self._api_client.has_iam_creds:
            return self._write_dataset_manifest(create_request)

        query = '''
          mutation CreateIngestDataset($input: CreateIngestDatasetInput!) {
            createIngestDataset(input: $input) {
              datasetId
              dataPath
            }
          }
        '''
        variables = {'input': create_request.to_json()}
        data: DatasetCreateResponse = self._api_client.query(query, variables=variables)['createIngestDataset']
        logger.info(f"Dataset ID: {data['datasetId']}")
        return data

    def get_dataset_files(self, project_id: str, dataset_id: str) -> List[File]:
        """
        Returns a list of file names that are in the provided dataset
        """
        access_context = FileAccessContext.download_dataset(project_id=project_id, dataset_id=dataset_id)
        return self._file_service.get_file_listing(access_context)

    def upload_files(self, project_id: str, dataset_id: str, directory: str, files: List[str]):
        """
        Uploads files to a given dataset from the specified local directory
        """
        access_context = FileAccessContext.upload_dataset(project_id=project_id, dataset_id=dataset_id)
        self._file_service.upload_files(access_context, directory, files)

    def download_files(self, project_id: str, dataset_id: str, download_location: str,
                       files: Union[List[File], List[str]] = None):
        """
         Downloads all the dataset files
         If the files argument isn't provided, all files will be downloaded
        """
        access_context = FileAccessContext.download_dataset(project_id=project_id, dataset_id=dataset_id)
        if files is None:
            files = self._file_service.get_file_listing(access_context)

        if len(files) == 0:
            return

        if isinstance(files[0], File):
            files = [file.relative_path for file in files]

        self._file_service.download_files(access_context, download_location, files)

    def _write_dataset_manifest(self, request: CreateIngestDatasetInput) -> DatasetCreateResponse:
        """
         Internal method for registering a dataset without API access.
         To be used for machine or service accounts
        """
        manifest = {
            'project': request.project_id,
            'process': request.process_id,
            'name': request.name,
            'desc': request.description,
            'infoJson': {
                'ingestedBy': self._api_client.current_user
            },
            'files': [{'name': file} for file in request.files]
        }
        dataset_id = str(uuid.uuid4())
        manifest_path = f'datasets/{dataset_id}/artifacts/manifest.json'
        manifest_json = json.dumps(manifest, indent=4)
        access_context = FileAccessContext.upload_dataset(dataset_id=dataset_id,
                                                          project_id=request.project_id)
        self._file_service.create_file(access_context,
                                       key=manifest_path,
                                       contents=manifest_json,
                                       content_type='application/json')
        return {
            'datasetId': dataset_id,
            'dataPath': f'datasets/{dataset_id}/artifacts/data'
        }



from cirro.api.models.reference import ReferenceType
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset


class DataPortalReferenceType(DataPortalAsset):
    """
    Reference data is organized by project, categorized by type.
    """
    name = None

    def __init__(self, ref_type: ReferenceType):
        self.name = ref_type.name
        self.description = ref_type.description
        self.directory = ref_type.directory
        self.validation = ref_type.validation

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'description']
        ])


class DataPortalReferenceTypes(DataPortalAssets[DataPortalReferenceType]):
    """Collection of DataPortalReferenceType objects."""
    asset_name = "reference type"

    def get_by_id(self, _id: str) -> DataPortalReferenceType:
        raise NotImplementedError("Filtering by ID is not supported, use get_by_name")



from abc import ABC, abstractmethod
from typing import Callable

from requests.auth import AuthBase


class AuthInfo(ABC):
    @abstractmethod
    def get_request_auth(self) -> AuthBase:
        raise NotImplementedError()

    @abstractmethod
    def get_current_user(self) -> str:
        raise NotImplementedError()


class RequestAuthWrapper(AuthBase):
    def __init__(self, token_getter: Callable[..., str]):
        self.token_getter = token_getter

    def __call__(self, request):
        request.headers['Authorization'] = self.token_getter()
        return request



from typing import List

from cirro.api.models.reference import ReferenceType
from cirro.api.services.base import BaseService


class CommonService(BaseService):
    def get_references_types(self) -> List[ReferenceType]:
        """
        Gets a list of available reference types
        """
        query = '''
          query GetReferenceTypes {
            getReferenceTypes {
              name
              description
              directory
              validation
            }
          }
        '''
        resp = self._api_client.query(query)['getReferenceTypes']
        return [ReferenceType.from_record(record) for record in resp]

    # TODO: Global reference catalogue



import fnmatch
from abc import abstractmethod
from typing import List, TypeVar

from cirro.sdk.exceptions import DataPortalAssetNotFound, DataPortalInputError


class DataPortalAsset:
    """Base class used for all Data Portal Assets"""

    @property
    @abstractmethod
    def name(self):
        pass


T = TypeVar('T', bound=DataPortalAsset)


class DataPortalAssets(List[T]):
    """
    Generic class with helper functions for any group of assets (projects, datasets, etc.) in the Data Portal.
    """

    # Overridden by child classes
    asset_name = 'asset'

    def __init__(self, input_list: List[T]):
        super().__init__(input_list)

    def __str__(self):
        return "\n\n".join([str(i) for i in self])

    def description(self):
        """Render a text summary of the assets."""

        return '\n\n---\n\n'.join([
            str(i)
            for i in self
        ])

    def get_by_name(self, name: str) -> T:
        """Return the item which matches with name attribute."""

        if name is None:
            raise DataPortalInputError(f"Must provide name to identify {self.asset_name}")

        # Get the items which have a matching name
        matching_queries = [i for i in self if i.name == name]

        # Error if no items are found
        msg = '\n'.join([f"No {self.asset_name} found with name '{name}'.", self.description()])
        if len(matching_queries) == 0:
            raise DataPortalAssetNotFound(msg)

        # Error if multiple projects are found
        msg = f"Multiple {self.asset_name} items found with name '{name}', use ID instead.\n{self.description()}"
        if len(matching_queries) > 1:
            raise DataPortalAssetNotFound(msg)

        return matching_queries[0]

    def get_by_id(self, _id: str) -> T:
        """Return the item which matches by id attribute."""

        if _id is None:
            raise DataPortalInputError(f"Must provide id to identify {self.asset_name}")

        # Get the items which have a matching ID
        matching_queries = [i for i in self if i.id == _id]

        # Error if no items are found
        msg = '\n'.join([f"No {self.asset_name} found with id '{_id}'.", self.description()])
        if len(matching_queries) == 0:
            raise DataPortalAssetNotFound(msg)

        return matching_queries[0]

    def filter_by_pattern(self, pattern: str) -> 'DataPortalAssets[T]':
        """Filter the items to just those whose name attribute matches the pattern."""

        # Get a list of the names to search against
        all_names = [i.name for i in self]

        # Filter the names by the pattern
        filtered_names = fnmatch.filter(all_names, pattern)

        # Filter the list to only include those items
        return self.__class__(
            [
                i
                for i in self
                if i.name in filtered_names
            ]
        )



import sys
from pathlib import Path
from typing import List

from prompt_toolkit.shortcuts import CompleteStyle
from prompt_toolkit.validation import Validator, ValidationError

from cirro.api.models.process import Process
from cirro.api.models.project import Project
from cirro.cli.interactive.common_args import ask_project
from cirro.cli.interactive.utils import prompt_wrapper
from cirro.cli.models import UploadArguments
from cirro.file_utils import get_directory_stats


class DataDirectoryValidator(Validator):
    def validate(self, document):
        is_a_directory = Path(document.text).is_dir()
        if not is_a_directory or len(document.text.strip()) == 0:
            raise ValidationError(
                message='Please enter a valid directory',
                cursor_position=len(document.text)
            )


def ask_data_directory(input_value: str) -> str:
    directory_prompt = {
        'type': 'path',
        'name': 'data_directory',
        'message': 'Enter the full path of the data directory',
        'validate': DataDirectoryValidator,
        'default': input_value or '',
        'complete_style': CompleteStyle.READLINE_LIKE,
        'only_directories': True
    }

    answers = prompt_wrapper(directory_prompt)
    return answers['data_directory']


def confirm_data_directory(directory: str):
    stats = get_directory_stats(directory)
    answers = prompt_wrapper({
        'type': 'confirm',
        'message': f'Please confirm that you wish to upload {stats["numberOfFiles"]} files ({stats["sizeFriendly"]})',
        'name': 'continue',
        'default': True
    })

    if not answers['continue']:
        sys.exit(1)


def ask_name(input_value: str) -> str:
    name_prompt = {
        'type': 'input',
        'name': 'name',
        'message': 'What is the name of this dataset?',
        'validate': lambda val: len(val.strip()) > 0 or 'Please enter a name',
        'default': input_value or ''
    }

    answers = prompt_wrapper(name_prompt)
    return answers['name']


def ask_description(input_value: str) -> str:
    description_prompt = {
        'type': 'input',
        'name': 'description',
        'message': 'Enter a description of the dataset (optional)',
        'default': input_value or ''
    }

    answers = prompt_wrapper(description_prompt)
    return answers['description']


def ask_process(processes: List[Process], input_value: str) -> str:
    process_names = [process.name for process in processes]
    process_prompt = {
        'type': 'list',
        'name': 'process',
        'message': 'What type of files?',
        'choices': process_names,
        'default': input_value if input_value in process_names else None
    }
    answers = prompt_wrapper(process_prompt)
    return answers['process']


def gather_upload_arguments(input_params: UploadArguments, projects: List[Project], processes: List[Process]):
    input_params['project'] = ask_project(projects, input_params.get('project'))

    input_params['data_directory'] = ask_data_directory(input_params.get('data_directory'))
    confirm_data_directory(input_params['data_directory'])

    input_params['process'] = ask_process(processes, input_params.get('process'))

    data_directory_name = Path(input_params['data_directory']).name
    default_name = input_params.get('name') or data_directory_name
    input_params['name'] = ask_name(default_name)
    input_params['description'] = ask_description(input_params.get('description'))
    return input_params



from typing import Optional

from cirro.api.auth import AuthInfo, get_auth_info_from_config
from cirro.api.clients import ApiClient
from cirro.api.config import AppConfig
from cirro.api.services import DatasetService, ProcessService, ProjectService, FileService, CommonService


class DataPortalClient:
    """
    A client for interacting with the Cirro platform
    """
    def __init__(self, auth_info: Optional[AuthInfo] = None, base_url: str = None):
        self._configuration = AppConfig(base_url=base_url)
        if not auth_info:
            auth_info = get_auth_info_from_config(self._configuration)

        self._api_client = ApiClient(auth_info, data_endpoint=self._configuration.data_endpoint)
        self._file_service = FileService(self._api_client, self._configuration)
        self._dataset_service = DatasetService(self._api_client, self._file_service, self._configuration)
        self._project_service = ProjectService(self._api_client, self._file_service, self._configuration)
        self._process_service = ProcessService(self._api_client, self._file_service, self._configuration)
        self._common_service = CommonService(self._api_client, self._configuration)

    @property
    def dataset(self):
        return self._dataset_service

    @property
    def project(self):
        return self._project_service

    @property
    def process(self):
        return self._process_service

    @property
    def file(self):
        return self._file_service

    @property
    def common(self):
        return self._common_service



import boto3
from requests.auth import AuthBase
from requests_aws4auth import AWS4Auth

from cirro.api.auth.base import AuthInfo
from cirro.api.models.auth import Creds


class IAMAuth(AuthInfo):
    """
    Uses AWS access tokens for authentication
    Note: this does not with any API calls at the moment
    """
    def __init__(self, region: str, access_key: str, secret_key: str, token: str = None):
        self.region = region
        self.creds: Creds = {
            'AccessKeyId': access_key,
            'SecretAccessKey': secret_key,
            'SessionToken': token,
            'Expiration': None
        }

    @classmethod
    def load_current(cls):
        """
        Loads the current session's AWS credentials
        """
        current = boto3.Session().get_credentials().get_frozen_credentials()
        return cls(current.access_key, current.secret_key, current.token)

    def get_request_auth(self) -> AuthBase:
        return AWS4Auth(self.creds['AccessKeyId'],
                        self.creds['SecretAccessKey'],
                        self.region,
                        'appsync',
                        session_token=self.creds['SessionToken'])

    def get_current_user(self) -> str:
        sts_client = boto3.client('sts',
                                  aws_access_key_id=self.creds['AccessKeyId'],
                                  aws_secret_access_key=self.creds['SecretAccessKey'],
                                  aws_session_token=self.creds['SessionToken'])
        identity_arn = sts_client.get_caller_identity()['Arn']
        username = identity_arn.split('/')[-1]
        return f'iam-{username}'



from cirro.api.clients.portal import DataPortalClient
from cirro.api.config import UserConfig, save_user_config
from cirro.api.models.dataset import CreateIngestDatasetInput
from cirro.api.models.process import Executor
from cirro.cli.interactive.auth_args import gather_auth_config
from cirro.cli.interactive.download_args import gather_download_arguments, ask_dataset_files
from cirro.cli.interactive.download_args import gather_download_arguments_dataset
from cirro.cli.interactive.list_dataset_args import gather_list_arguments
from cirro.cli.interactive.upload_args import gather_upload_arguments
from cirro.cli.interactive.utils import get_id_from_name, get_item_from_name_or_id
from cirro.cli.interactive.workflow_args import get_preprocess_script, get_additional_inputs, get_outputs, \
    get_child_processes, \
    get_repository, get_description, get_output_resources_path
from cirro.cli.interactive.workflow_form_args import prompt_user_inputs, get_nextflow_schema, convert_nf_schema
from cirro.cli.models import ListArguments, UploadArguments, DownloadArguments
from cirro.file_utils import get_files_in_directory
from cirro.helpers import WorkflowConfigBuilder


def run_list_datasets(input_params: ListArguments, interactive=False):
    """List the datasets available in a particular project."""

    # Instantiate the Cirro Data Portal client
    cirro = DataPortalClient()

    # If the user provided the --interactive flag
    if interactive:

        # Get the list of projects available to the user
        projects = cirro.project.list()

        if len(projects) == 0:
            print("No projects available")
            return

        # Prompt the user for the project
        input_params = gather_list_arguments(input_params, projects)

    # List the datasets available in that project
    datasets = cirro.dataset.find_by_project(input_params['project'])

    sorted_datasets = sorted(datasets, key=lambda d: d.created_at, reverse=True)
    print("\n\n".join([f'Name: {dataset.name}\n'
                       f'Desc: {dataset.description}\n'
                       f'GUID: ({dataset.id})'
                       for dataset in sorted_datasets]))


def run_ingest(input_params: UploadArguments, interactive=False):
    cirro = DataPortalClient()
    projects = cirro.project.list()
    processes = cirro.process.list(process_type=Executor.INGEST)

    if len(projects) == 0:
        print("No projects available")
        return

    if interactive:
        input_params = gather_upload_arguments(input_params, projects, processes)

    directory = input_params['data_directory']
    files = get_files_in_directory(directory)
    if len(files) == 0:
        raise RuntimeWarning("No files to upload, exiting")

    process = get_item_from_name_or_id(processes, input_params['process'])
    cirro.process.check_dataset_files(files, process.id, directory)

    create_request = CreateIngestDatasetInput(
        project_id=get_id_from_name(projects, input_params['project']),
        process_id=process.id,
        name=input_params['name'],
        description=input_params['description'],
        files=files
    )

    create_resp = cirro.dataset.create(create_request)

    cirro.dataset.upload_files(dataset_id=create_resp['datasetId'],
                               project_id=create_request.project_id,
                               directory=directory,
                               files=files)


def run_download(input_params: DownloadArguments, interactive=False):
    cirro = DataPortalClient()

    projects = cirro.project.list()

    if len(projects) == 0:
        print("No projects available")
        return

    files_to_download = None
    if interactive:
        input_params = gather_download_arguments(input_params, projects)

        input_params['project'] = get_id_from_name(projects, input_params['project'])
        datasets = cirro.dataset.find_by_project(input_params['project'])
        input_params = gather_download_arguments_dataset(input_params, datasets)
        dataset_files = cirro.dataset.get_dataset_files(input_params['project'], input_params['dataset'])
        files_to_download = ask_dataset_files(dataset_files)

    dataset_params = {
        'project': get_id_from_name(projects, input_params['project']),
        'dataset': input_params['dataset']
    }

    cirro.dataset.download_files(project_id=dataset_params['project'],
                                 dataset_id=dataset_params['dataset'],
                                 download_location=input_params['data_directory'],
                                 files=files_to_download)


def run_configure_workflow():
    """Configure a workflow to be run in the Data Portal as a process."""

    cirro = DataPortalClient()
    process_options = cirro.process.list(process_type=Executor.NEXTFLOW)
    resources_folder, repo_prefix = get_output_resources_path()

    workflow = WorkflowConfigBuilder(repo_prefix)

    # Process record
    repo = get_repository()
    workflow.with_repository(repo)

    # Prompt for optional pre-process script
    if (preprocess_py := get_preprocess_script()) is not None:
        workflow.with_preprocess(preprocess_py)

    workflow.with_child_processes(
        get_child_processes(process_options)
    )

    # Process compute
    workflow.with_compute()

    # Process form & process inputs
    nf_schema = get_nextflow_schema(repo.repo_path, repo.version)
    inputs = {}
    if nf_schema is not None:
        form = {**nf_schema}
        convert_nf_schema(form, inputs)

        workflow.with_description(nf_schema['description'])
        workflow.with_form_inputs(form)

    else:
        workflow.with_description(get_description())
        form, inputs = prompt_user_inputs()
        workflow.with_form_inputs(form)

    # Inputs based on form config
    workflow.with_inputs(inputs)

    # Additional process inputs
    for input_name, input_value in get_additional_inputs():
        workflow.with_input(input_name, input_value)

    # Process outputs
    for output in get_outputs():
        workflow.with_output(output)
    workflow.with_common_outputs()

    # Save to resources
    workflow.save_local(resources_folder)


def run_configure():
    auth_method, auth_method_config = gather_auth_config()
    save_user_config(UserConfig(auth_method=auth_method,
                                auth_method_config=auth_method_config,
                                base_url=None))



from typing import List

from cirro.api.models.project import Project
from cirro.cli.interactive.utils import prompt_wrapper


def ask_project(projects: List[Project], input_value: str) -> str:
    project_names = [project.name for project in projects]
    project_prompt = {
        'type': 'list',
        'name': 'project',
        'message': 'What project is this dataset associated with?',
        'choices': project_names,
        'default': input_value if input_value in project_names else None
    }
    answers = prompt_wrapper(project_prompt)
    return answers['project']



from typing import List, Optional

from cirro.api.clients.utils import filter_deleted
from cirro.api.models.file import FileAccessContext
from cirro.api.models.project import Project
from cirro.api.models.reference import Reference, References
from cirro.api.services.file import FileEnabledService
from cirro.file_utils import filter_files_by_pattern


class ProjectService(FileEnabledService):
    def list(self) -> List[Project]:
        """
        Gets a list of projects that you have access to
        """
        query = '''
          query ListProjects(
            $filter: ModelProjectFilterInput
            $limit: Int
            $nextToken: String
          ) {
            listProjects(filter: $filter, limit: $limit, nextToken: $nextToken) {
              items {
                id
                name
                desc
                status
              }
            }
          }
        '''

        resp = self._api_client.query(query)['listProjects']
        items = filter_deleted(resp['items'])
        return [Project.from_record(item) for item in items]

    def find_by_name(self, name: str) -> Optional[Project]:
        """
        Finds a project by name
        """
        return next((p for p in self.list() if p.name == name), None)

    def get_reference_types(self, project_id: str) -> List[str]:
        """
        Gets the list of reference types which are available in a project
        """
        ref_prefix = "data/references/"
        access_context = FileAccessContext.download_project_resources(project_id)
        resources = self._file_service.get_file_listing(access_context)
        reference_files = filter_files_by_pattern(resources, f'{ref_prefix}*/*/*')
        return list(set([file.relative_path[len(ref_prefix):].split("/", 1)[0] for file in reference_files]))

    def get_references(self, project_id: str, reference_directory: str) -> References:
        """
        Gets a list of references available for a given project and reference directory
        """
        access_context = FileAccessContext.download_project_resources(project_id)
        resources = self._file_service.get_file_listing(access_context)
        reference_files = filter_files_by_pattern(resources, f'data/references/{reference_directory}/*/*')
        references = [*set(Reference.of(file) for file in reference_files)]
        references = sorted(references, key=lambda r: r.name.lower())
        return References(references)



from cirro.api.clients.portal import DataPortalClient
from cirro.api.models.form_specification import ParameterSpecification
from cirro.api.models.process import Process
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset


class DataPortalProcess(DataPortalAsset):
    """Helper functions for interacting with analysis processes."""
    name = None

    def __init__(self, process: Process, client: DataPortalClient):

        self.id = process.id
        self.name = process.name
        self.description = process.description
        self.child_process_ids = process.child_process_ids
        self.executor = process.executor
        self.documentation_url = process.documentation_url
        self.code = process.code
        self.form_spec_json = process.form_spec_json
        self.sample_sheet_path = process.sample_sheet_path
        self.file_requirements_message = process.file_requirements_message
        self.file_mapping_rules = process.file_mapping_rules
        self._client = client

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description']
        ])

    def get_parameter_spec(self) -> ParameterSpecification:
        """
        Gets a specification used to describe the parameters used in the process.
        """
        return self._client.process.get_parameter_spec(self.id)


class DataPortalProcesses(DataPortalAssets[DataPortalProcess]):
    """Collection of DataPortalProcess objects."""
    asset_name = "process"
